import numpy as np
import pandas as pd
from pathlib import Path
from pydantic import validate_arguments


@validate_arguments
def kmm(path: Path):
    return pd.read_csv(
        path,
        sep="\t",
        encoding="latin1",
        names=[
            "centimeter",
            "track_section",
            "kilometer",
            "meter",
            "track_lane",
            "1?",
            "2?",
            "3?",
            "4?",
            "5?",
            "sweref99_tm_x",
            "sweref99_tm_y",
            "8?",
            "9?",
        ],
        dtype=dict(
            track_section=str,
            kilometer=np.int32,
            meter=np.int32,
            track_lane=str,
            sweref99_tm_x=np.float32,
            sweref99_tm_y=np.float32,
        ),
    )



import re
from pathlib import Path
import pandas as pd
from xml.etree import ElementTree
from pydantic import validate_arguments


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def camera_positions(positions: pd.DataFrame, header_path: Path):
    ascending = kmm_ascending(positions)
    direction = car_direction(header_path)

    if (direction == "A" and ascending) or (direction == "B" and not ascending):
        correction = -8

    elif (direction == "A" and not ascending) or (direction == "B" and ascending):
        correction = 8

    else:
        raise ValueError

    return (
        positions
        .assign(meter=lambda df: df["meter"] + correction)
    )


def kmm_ascending(positions):
    total_meter = positions["kilometer"] * 1000 + positions["meter"]
    diff = total_meter.values[:-1] - total_meter.values[1:]
    descending = (diff < 0).mean()
    ascending = (diff > 0).mean()

    if descending < 0.9 and ascending < 0.9:
        raise ValueError("Unable to determine ascending/descending kmm numbers")

    else:
        return ascending > 0.9


def car_direction(header_path: Path):

    tree = ElementTree.parse(header_path)
    root = tree.getroot()
    start_tags = [
        child.text
        for child in root
        if child.tag == "Start"
    ]

    if len(start_tags) != 1:
        raise ValueError(f"Expected 1 'Start' tag in header file, found {len(start_tags)}")

    start_tag = start_tags[0]
    car_direction = re.search(
        r"CarDirection = \"(.*)\"",
        start_tag,
    ).group(1)

    if car_direction not in ["A", "B"]:
        raise ValueError(f"Unkown car direction {car_direction}")

    return car_direction


def test_camera_positions_kmm():
    from kmm.kmm import kmm

    df = kmm("tests/ascending_B.kmm")
    df_calibrated = camera_positions(df, "tests/ascending_B.hdr")
    assert df_calibrated["meter"].iloc[0] == df["meter"].iloc[0] - 8


def test_camera_positions_kmm2():
    from kmm.kmm2 import kmm2

    df = kmm2("tests/ascending_B.kmm2")
    df_calibrated = camera_positions(df, "tests/ascending_B.hdr")
    assert df_calibrated["meter"].iloc[0] == df["meter"].iloc[0] - 8



from kmm.kmm import kmm
from kmm.kmm2 import kmm2
from kmm.read_raw import read_raw
from kmm.geodetic import geodetic
from kmm.camera_positions import camera_positions
from kmm.sync_frame_index import sync_frame_index
from kmm.position_adjustment import PositionAdjustment
from kmm.read_sync_adjust import read_sync_adjust

WIRE_CAMERA_ADJUSTMENT = 1



import pandas as pd
from sweref99 import projections


tm = projections.make_transverse_mercator("SWEREF_99_TM")


def geodetic(positions: pd.DataFrame):
    if len(positions) == 0:
        positions = positions.assign(longitude=[], latitude=[])
    else:
        latitude, longitude = zip(*positions.apply(
            lambda row: (tm.grid_to_geodetic(row["sweref99_tm_x"], row["sweref99_tm_y"])),
            axis="columns",
            result_type="reduce",
        ))
        positions = positions.assign(longitude=longitude, latitude=latitude)
    return positions


def test_geodetic():
    from kmm.kmm2 import kmm2

    df = kmm2("tests/ascending_B.kmm2")
    df = geodetic(df)
    assert ((df["latitude"] < 68) & (df["latitude"] > 55)).all()
    assert ((df["longitude"] < 25) & (df["longitude"] > 7)).all()



import re
import pandas as pd
from pathlib import Path
from xml.etree import ElementTree
from pydantic import validate_arguments


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def sync_frame_index(positions: pd.DataFrame, header_path: Path):

    tree = ElementTree.parse(header_path)
    root = tree.getroot()
    sync_tags = [
        child.text
        for child in root
        if child.tag == "Sync"
    ]

    if len(sync_tags) == 0:
        raise ValueError("Did not find a sync tag.")

    sync_tag = sync_tags[0]

    position = int(re.search(
        r"Position = \"(\d*)\"",
        sync_tag,
    ).group(1))

    sync = int(re.search(
        r"Sync = \"(\d*)\"",
        sync_tag,
    ).group(1))

    return positions.assign(
        frame_index=((positions["centimeter"] + position - sync) / 10).astype(int)
    )


def test_sync_frame_index():
    from kmm.kmm2 import kmm2

    df = kmm2("tests/ascending_B.kmm2")
    df = sync_frame_index(df, "tests/ascending_B.hdr")
    assert (~df["frame_index"].isnull()).all()



from pathlib import Path
from pydantic import validate_arguments

import kmm


@validate_arguments
def read_sync_adjust(
    path: Path,
    header_path: Path,
    adjust: int = kmm.PositionAdjustment.WIRE_CAMERA,
    geodetic: bool = True,
):
    """
    Used to load a kmm or kmm2 file as a pandas DataFrame.

    Example:

    .. code-block:: python

        from pathlib import Path
        import kmm

        path = Path("...")
        header_path = Path("...")

        df = kmm.read(path, header_path)

    """

    df = kmm.read_raw(path)
    df = kmm.sync_frame_index(df, header_path)
    if adjust == kmm.PositionAdjustment.WIRE_CAMERA:
        df = kmm.camera_positions(df, header_path)
    if geodetic:
        df = kmm.geodetic(df)

    return df


def test_read_kmm():
    df = read_sync_adjust("tests/ascending_B.kmm", "tests/ascending_B.hdr")
    assert len(df) > 0


def test_read_kmm2():
    df = read_sync_adjust("tests/ascending_B.kmm2", "tests/ascending_B.hdr")
    assert len(df) > 0


def test_empty_kmm():
    df = read_sync_adjust("tests/empty.kmm", "tests/empty.hdr")
    assert len(df) == 0


def test_empty_kmm2():
    df = read_sync_adjust("tests/empty.kmm2", "tests/empty.hdr")
    assert len(df) == 0



import re
import numpy as np
import pandas as pd
from pathlib import Path
from pydantic import validate_arguments


pattern = re.compile(r".+\[.+\]")
pattern2 = re.compile(r"CMAST")


@validate_arguments
def kmm2(path: Path):

    skiprows = [
        index
        for index, line in enumerate(path.read_text(encoding="latin1").splitlines())
        if pattern.match(line) or pattern2.match(line)
    ]

    return pd.read_csv(
        path,
        skiprows=[0] + skiprows,
        delimiter="\t",
        encoding="latin1",
        names=[
            "code",
            "centimeter",
            "track_section",
            "kilometer",
            "meter",
            "track_lane",
            "1?",
            "2?",
            "3?",
            "4?",
            "sweref99_tm_x",
            "sweref99_tm_y",
            "contact_wire_material",
            "rail_model",
            "sliper_model",
            "between_stations",
            "5?",
            "6?",
            "7?",
            "8?",
            "max_speed",
        ],
        dtype=dict(
            centimeter=np.int64,
            track_section=str,
            kilometer=np.int32,
            meter=np.int32,
            track_lane=str,
            sweref99_tm_x=np.float32,
            sweref99_tm_y=np.float32,
        ),
        low_memory=False,
    )


def test_patterns():
    assert pattern.match("Västerås central [Vå]")
    assert pattern2.match("CMAST   281-2B")



from pathlib import Path
from pydantic import validate_arguments

import kmm


@validate_arguments
def read_raw(path: Path):
    if path.suffix == ".kmm":
        df = kmm.kmm(path)
    elif path.suffix == ".kmm2":
        df = kmm.kmm2(path)
    else:
        raise ValueError(f"Unable to parse file type {path.suffix}")

    return df
